# coding:utf-8
import getpass
import re
import os
import shutil
import stat
import subprocess as subs
import crypt
import csv
import logging

logging.basicConfig(level=logging.INFO)


def insert(file, file_data):
    flags = os.O_RDWR | os.O_CREAT
    mode = stat.S_IWUSR | stat.S_IRUSR

    if os.path.exists(file):
        os.remove(file)
    with os.fdopen(os.open(file, flags, mode), "w") as f:
        f.write(file_data)


# Replaces the string in the file
def alter(file, old_str, new_str):
    file_data = ""

    try:
        file_data = alter_insert(file, old_str, new_str)

    except FileNotFoundError:
        logging.info("The file does not exist")
    insert(file, file_data)


def alter_insert(file, old_str, new_str):
    file_data = ""
    with open(file, "r", encoding="utf-8") as f:
        for line in f:
            if old_str in line:
                line = line.replace(old_str, new_str)
            file_data += line
    return file_data


def alter_after(file, find_str, new_str):
    file_data = ""
    try:
        file_data = alter_after_insert(file, find_str, new_str)
    except FileNotFoundError:
        logging.info("The file does not exist")
        return
    insert(file, file_data)


def alter_after_insert(file, find_str, new_str):
    file_data = ''
    with open(file, "r", encoding="utf-8") as f:
        for line in f:
            if find_str in line:
                begin = line[:line.find(find_str)]
                line = begin + find_str + new_str
            file_data += line
    return file_data


def add_after(file, find_str, new_str):
    file_data = ""
    try:
        file_data = add_after_insert(file, find_str, new_str)
    except FileNotFoundError:
        logging.info("The file does not exist")
    insert(file, file_data)


def add_after_insert(file, find_str, new_str):
    file_data = ""
    with open(file, "r", encoding="utf-8") as f:
        for line in f:
            if find_str in line:
                file_data += line
                file_data += new_str
                continue
            file_data += line
    return file_data


def read_csv(filename):
    dic = {}
    li = []
    ip_map_v = []
    with open(filename, newline='') as csvf:
        reader = csv.reader(csvf)
        for row in reader:
            if row:
                li.append(row)

    for i in li:
        if i[0] == "ipMap":
            ip_map_v.append(i[1])
        dic.update({i[0]: i[1]})
    if ip_map_v:
        dic.update({"ipMap": ip_map_v})
    return dic


def change_ks(kick_start, ip, dic):
    file_data = """#version=DEVEL
# Use graphical install
graphical
#Use http installation source
url --url=http://ip/FusionOS/

%packages
@^server-product-environment
expect
%end

#add

# Root password
rootpw --iscrypted
"""

    router = dic.get("routers")
    command_syn("mkdir -m 755 /var/www/html/ks/")
    if "bond" in dic:
        cust_bond = dic.get("bond")
        if "ipMap" not in dic:
            logging.error("Unable to find 'ipMap' to configure bond, need add first ipMap")
            raise ValueError("Unable to find 'ipMap' to configure bond")

        shutil.copy("/etc/auto-pxe/bond_ks.cfg", kick_start)
        alter(kick_start, "192.1.1.1", router)
        nic_name = dic.get("nicName", None)
        if nic_name is not None and len(nic_name.strip()) > 0:
            nic1, nic2 = parser_nic_name(nic_name)
            command_syn(f"rm -f {kick_start}")
            shutil.copy("/etc/auto-pxe/static_bond.cfg", kick_start)
            alter(kick_start, "NIC1", nic1)
            alter(kick_start, "NIC2", nic2)
        if len(cust_bond.strip()) > 0:
            alter(kick_start, "bond1", cust_bond)
        bond_mode = dic.get("bondMode", None)
        if bond_mode and len(bond_mode.strip()) > 0:
            bond_mode_list = [
                "active-backup",
                "balance-rr",
                "balance-xor",
                "broadcast",
                "802.3ad",
                "balance-tlb",
                "balance-alb"
            ]
            if bond_mode not in bond_mode_list:
                logging.error(f"Unsupported mode: {bond_mode}")
                raise ValueError("Unsupported bond mode")
            alter(kick_start, "active-backup", bond_mode)

    elif dic.get("ipMap", None):
        shutil.copy("/etc/auto-pxe/static_ks.cfg", kick_start)
        alter(kick_start, "192.1.1.1", router)
    else:
        flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        if os.path.exists(kick_start):
            os.remove(kick_start)
        with os.fdopen(os.open(kick_start, flags, mode), "w") as f:
            f.write(file_data)

    while 1:
        password_input = getpass.getpass("Please enter root password you want： ")
        password_input2 = getpass.getpass("Please enter it again: ")
        if password_input == password_input2:
            logging.info("successfully")
            break
        else:
            logging.info("two inputs are inconsistent , please try again")
    password_output = crypt.crypt(password_input)
    alter_after(kick_start, "rootpw --iscrypted", " " + password_output)

    # Modify the configuration according to the csv file
    dic2 = read_csv('/etc/auto-pxe/ks_cfg.csv')
    s = "http://%s/FusionOS/\n" % ip
    alter_after(kick_start, "url --url=", s)

    # The following are custom configurations
    data_ks = ""
    for _, v in dic2.items():
        if v != "":
            data_ks += v
            data_ks += '\n'

    alter_after(kick_start, "add\n", data_ks)
    command_syn(f"chmod 644 {kick_start}")


def x86grub(ip, iso_name):
    ## x86 uefi and legacy support
    x86_uefi_base_dir = "/var/lib/tftpboot/x86_uefi"
    shutil.copytree("/mnt/images/pxeboot/", x86_uefi_base_dir)

    shutil.copy("/mnt/EFI/BOOT/BOOTX64.EFI", x86_uefi_base_dir)
    shutil.copy("/mnt/EFI/BOOT/grubx64.efi", x86_uefi_base_dir)
    shutil.copy("/mnt/EFI/BOOT/grub.cfg", x86_uefi_base_dir)

    grub = f"{x86_uefi_base_dir}/grub.cfg"
    alter(grub, '''set default="1"''', '''set default="0"''')
    s = " '%s' \n" % iso_name[:-4]
    alter_after(grub, "search --no-floppy --set=root -l", s)

    add_ip = "       set root=(tftp,%s)\n" % ip
    add_after(grub, "menuentry 'Install FusionOS 23' --class FusionOS \
        --class gnu-linux --class gnu --class os {", add_ip)
    add_after(grub, "menuentry 'Test this media & install FusionOS 23' \
        --class FusionOS --class gnu-linux --class gnu --class os {", add_ip)
    s = " quiet inst.ks=http://%s/ks/FusionOS-ks.cfg\n" % ip
    alter_after(grub, "linuxefi /images/pxeboot/vmlinuz", s)
    command_syn(f"chmod 444 {grub}")

    ## x86 legacy blos support
    x86_legacy_base_dir = "/var/lib/tftpboot/pxelinux"
    shutil.copytree("/mnt/images/pxeboot/", x86_legacy_base_dir)

    files = get_file_by_path("/mnt/Packages/")
    tmp_path = "/tmp/auto-pxe"
    command_syn(f"mkdir {tmp_path}")
    for file_name in files:
        if file_name.startswith("syslinux-nonlinux-") and file_name.endswith(".rpm"):
            cp_cmd = f"cp -pr /mnt/Packages/{file_name} {tmp_path}"
            command_syn(cp_cmd)
            rpm_cmd_1 = f"rpm2cpio {tmp_path}/{file_name}"
            rpm_cmd_2 = f"cpio -dimv -D {tmp_path}"
            retcode, output = command_link(rpm_cmd_1, rpm_cmd_2)
            logging.info(retcode)

    shutil.copy(f"{tmp_path}/usr/share/syslinux/pxelinux.0", x86_legacy_base_dir)
    shutil.copy(f"{tmp_path}/usr/share/syslinux/vesamenu.c32", x86_legacy_base_dir)
    shutil.copy(f"{tmp_path}/usr/share/syslinux/libcom32.c32", x86_legacy_base_dir)
    shutil.copy(f"{tmp_path}/usr/share/syslinux/libutil.c32", x86_legacy_base_dir)
    shutil.copy(f"{tmp_path}/usr/share/syslinux/ldlinux.c32", x86_legacy_base_dir)
    command_syn(f"rm -rf {tmp_path}")

    # blos config file, like grub.cfg
    command_syn(f"mkdir -m 755 -p {x86_legacy_base_dir}/pxelinux.cfg")
    pxelinux_default_config = f"""
default vesamenu.c32
prompt 1
timeout 60

display boot.msg

label linux
  menu label ^Install system
  menu default
  kernel vmlinuz
  append initrd=initrd.img inst.ks=http://{ip}/ks/FusionOS-ks.cfg ip=dhcp inst.repo=http://{ip}/FusionOS
label vesa
  menu label Install system with ^basic video drive
  kernel vmlinuz
  append initrd=initrd.img inst.ks=http://{ip}/ks/FusionOS-ks.cfg ip=dhcp inst.xdriver=vesa nomodeset inst.repo=http://{ip}/FusionOS
label rescue
  menu label ^Rescue installed system
  kernel vmlinuz
  append initrd=initrd.img rescue
label local
  menu label Boot from ^local drive
  localboot 0xffff
"""

    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
    mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
    ks_path = "/var/lib/tftpboot/pxelinux/pxelinux.cfg/default"
    if os.path.exists(ks_path):
        os.remove(ks_path)
    with os.fdopen(os.open(ks_path, flags, mode), "w") as f:
        f.write(pxelinux_default_config)

    ## vmlinuz and initrd.img
    command_syn("mkdir -m 755 -p /var/lib/tftpboot/images/pxeboot")
    shutil.copy(f"{x86_uefi_base_dir}/vmlinuz", "/var/lib/tftpboot/images/pxeboot")
    shutil.copy(f"{x86_uefi_base_dir}/initrd.img", "/var/lib/tftpboot/images/pxeboot")
    chmod_by_file("/var/lib/tftpboot/images")


def armgrub(ip, iso_name):
    arm_uefi_base_dir = "/var/lib/tftpboot/arm_uefi"

    shutil.copytree("/mnt/images/pxeboot/", arm_uefi_base_dir)

    shutil.copy("/mnt/EFI/BOOT/BOOTAA64.EFI", arm_uefi_base_dir)
    shutil.copy("/mnt/EFI/BOOT/grubaa64.efi", arm_uefi_base_dir)
    shutil.copy("/mnt/EFI/BOOT/grub.cfg", arm_uefi_base_dir)

    command_syn("mkdir -m 755 -p /var/lib/tftpboot/images/pxeboot")
    shutil.copy(f"{arm_uefi_base_dir}/vmlinuz", "/var/lib/tftpboot/images/pxeboot")
    shutil.copy(f"{arm_uefi_base_dir}/initrd.img", "/var/lib/tftpboot/images/pxeboot")

    chmod_by_file("/var/lib/tftpboot/images")

    grub = f"{arm_uefi_base_dir}/grub.cfg"
    alter(grub, '''set default="1"''', '''set default="0"''')
    s = " %s \n" % iso_name[:-4]
    alter_after(grub, "search --no-floppy --set=root -l", s)
    s = "@60me inst.ks=http://%s/ks/FusionOS-ks.cfg\n" % ip
    alter_after(grub, "@60me", s)

    alter(grub, "inst.stage2=hd:LABEL=FusionOS-23-aarch64", "")
    command_syn(f"chmod 444 {grub}")


def add_yum_from_iso():
    iso = "/home/FusionOS"
    file_names = get_file_by_path(iso)
    iso_name = ""
    tar_name = ""
    iso_num, tar_num = 0, 0
    for s in file_names:
        if s[-3:] == "iso":
            iso_name = s
            iso_num += 1
        if s[-2:] == "gz":
            tar_name = s
            tar_num += 1
    if iso_num == 0:
        logging.info("iso does not exist in the /home/FusionOS, is the path wrong or forgot to import?")
    if tar_num == 0:
        logging.info("The tar package does not exist in the /home/FusionOS, is it a wrong path or forgot to import?")
    if iso_num > 1:
        logging.info("There are multiple iso, the last one will be selected as the default")
    if tar_num > 1:
        logging.info("There are multiple tar, the last one will be selected as the default")

    if iso_name[-11:-4] == "aarch64":
        logging.info("After judgment, the server version is arm64")
    else:
        logging.info("After judgment, the server version is x86")
    chmod_by_file('/mnt')
    chmod_by_file('/home/FusionOS/')
    s = "mount /home/FusionOS/%s /mnt/" % iso_name
    command_syn(s)

    s2 = 'mkdir -m 755 -p /home/FusionOS/srv/repo/'
    s3 = 'cp -r /mnt/Packages /home/FusionOS/srv/repo/'
    s4 = 'cp -r /mnt/repodata /home/FusionOS/srv/repo'
    s5 = 'cp -r /mnt/RPM-GPG-KEY-FusionOS /home/FusionOS/srv/repo/'
    s = 'mount /home/FusionOS/%s /mnt/' % iso_name
    command_syn(s)
    command_syn(s2)
    command_syn(s3)
    command_syn(s4)
    command_syn(s5)
    file = "/etc/yum.repos.d/FusionOS.repo"
    file_data = '''[base]
name=base
baseurl=file:///home/FusionOS/srv/repo
enabled=1
gpgcheck=1
gpgkey=file:///home/FusionOS/srv/repo/RPM-GPG-KEY-FusionOS
[update]
name=update
baseurl=file:///home/FusionOS/srv/update/
enabled=0
gpgcheck=1
gpgkey=file:///home/FusionOS/srv/repo/RPM-GPG-KEY-FusionOS '''

    insert(file, file_data)
    command_syn('mkdir -m 755 -p /home/FusionOS/srv/update')
    s = 'tar -xf /home/FusionOS/%s -C /home/FusionOS/srv/update' % tar_name
    command_syn(s)
    command_syn('dnf install createrepo -y')
    command_syn('createrepo /home/FusionOS/srv/update')

    alter(file, "enabled=0", "enabled=1")


# if client is in secureboot mode, modify filename from grubx64.efi(grubaa64.efi) to BOOTX64.EFI(BOOTAA64.EFI)
def add_dhcp(ip, dic):
    command_syn("dnf install dhcp -y")
    dhcpd = "/etc/dhcp/dhcpd.conf"
    shutil.copy("/usr/share/doc/dhcp-server/dhcpd.conf.example", dhcpd)
    chmod_by_file(dhcpd)

    file = '''# DHCP Server Configuration file.
#   see /usr/share/doc/dhcp-server/dhcpd.conf.example
#   see dhcpd.conf(5) man page
#
option space pxelinux;
option pxelinux.magic code 208 = string;
option pxelinux.configfile code 209 = text;
option pxelinux.pathprefix code 210 = text;
option pxelinux.reboottime code 211 = unsigned integer 32;
option architecture-type code 93 = unsigned integer 16;
ddns-update-style interim;
ignore client-updates;

subnet 192.168.122.0 netmask 255.255.255.0 {
    option routers 192.168.122.1;
    option subnet-mask 255.255.255.0;
    range dynamic-bootp 192.168.122.50 192.168.122.200;
    default-lease-time 21600;
    max-lease-time 43200;
    if option architecture-type = 00:07 or option architecture-type = 00:09 {
      filename "x86_uefi/grubx64.efi";
    } else if option architecture-type = 00:0b {
      filename "arm_uefi/grubaa64.efi";
    } else{
      filename "pxelinux/pxelinux.0";
    }
    next-server 192.168.122.1;
}'''

    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
    mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
    if os.path.exists(dhcpd):
        os.remove(dhcpd)
    with os.fdopen(os.open(dhcpd, flags, mode), "w") as f:
        f.write(file)

    routers = dic["routers"]
    netmask = dic["netmask"]
    dynamic_bootp = dic["dynamicBootp"]
    subnet = dic["subnet"]
    alter(dhcpd, "192.168.122.0", subnet)
    alter(dhcpd, "255.255.255.0", netmask)
    line_break = ";\n"
    alter_after(dhcpd, "option routers ", routers + line_break)
    alter_after(dhcpd, "option subnet-mask ", netmask + line_break)
    alter_after(dhcpd, "range dynamic-bootp ", dynamic_bootp + line_break)
    alter_after(dhcpd, "next-server ", ip + line_break)
    command_syn("systemctl start dhcpd")
    command_syn("systemctl enable dhcpd")


def judge_ip(ip):
    compile_ip = re.compile('^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$')
    if compile_ip.match(ip):
        return True
    else:
        return False


def judge_mac(mac):
    compile_ip = re.compile('^([0-9a-f]{2}(?::[0-9a-f]{2}){5})$')
    if compile_ip.match(mac.lower()):
        return True
    else:
        return False


def parser_ip_map(ip_map_list: list):
    ip_map = {}
    ip_tuple = []
    for i in ip_map_list:
        try:
            mac, ip = i.split(" ")
        except Exception as e:
            logging.error(f"parser ipMap error, value is: {i}")
            raise ValueError("parser ipMap error") from e
        if not judge_ip(ip):
            logging.error(f"{ip} is not in ip format")
            raise ValueError(f"parser ipMap error, value is: {i}")
        if not judge_mac(mac):
            logging.error(f"{mac} is not in mac format")
            raise ValueError(f"parser ipMap error, value is: {i}")
        if mac in ip_map:
            logging.error(f"{mac} mac address already exists")
            raise ValueError(f"parser ipMap error, value is: {i}")
        if ip in ip_tuple:
            logging.error(f"{ip} ip address already exists")
            raise ValueError(f"parser ipMap error, value is: {i}")
        ip_tuple.append(ip)
        ip_map[mac] = ip
    return ip_map


def parser_nic_name(nic_name):
    try:
        nic1, nic2 = nic_name.split(" ")
    except Exception as e:
        logging.error(f"parser nicname error, value is: {nic_name}")
        raise ValueError("parser ipMap error") from e
    return nic1, nic2


def parser_dynamic(dynamic):
    try:
        begin, end = dynamic.split(" ")
    except Exception as e:
        logging.error(f"parser dynamic error, value is: {dynamic}")
        raise ValueError("parser dynamic error") from e
    if not judge_ip(begin):
        logging.error(f"{begin} is not in ip format")
        raise ValueError(f"parser ipMap error")
    if not judge_ip(end):
        logging.error(f"{end} is not in ip format")
        raise ValueError(f"parser ipMap error")
    if begin == end:
        logging.error("dynamicBootp scope is wrong")
        raise ValueError(f"parser ipMap error")


def run_dhcp(ip, version):
    dic = read_csv("/etc/auto-pxe/network.csv")
    logging.info("#find ipMap value, now add fixed ip")
    try:
        routers = dic["routers"]
        netmask = dic["netmask"]
        dynamic_bootp = dic["dynamicBootp"]
        subnet = dic["subnet"]
        ip_map_list = dic.get("ipMap", None)
    except KeyError:
        logging.info("Network configuration missing entries, please fill in")
    for i in [routers, netmask, subnet]:
        if not judge_ip(i):
            logging.error(f"{i}  is wrong format")
            raise ValueError(f"parser ipMap error")
    parser_dynamic(dynamic_bootp)
    if ip_map_list is not None:
        add_dhcp_static(ip, dic)
    else:
        add_dhcp(ip, dic)


def add_dhcp_static(ip, dic):
    command_syn("dnf install dhcp -y")
    dhcpd = "/etc/dhcp/dhcpd.conf"
    shutil.copy("/usr/share/doc/dhcp-server/dhcpd.conf.example", dhcpd)
    chmod_by_file(dhcpd)

    
    file = '''# DHCP Server Configuration file.
#   see /usr/share/doc/dhcp-server/dhcpd.conf.example
#   see dhcpd.conf(5) man page
#
option space pxelinux;
option pxelinux.magic code 208 = string;
option pxelinux.configfile code 209 = text;
option pxelinux.pathprefix code 210 = text;
option pxelinux.reboottime code 211 = unsigned integer 32;
option architecture-type code 93 = unsigned integer 16;
ddns-update-style interim;
ignore client-updates;
subnet 192.168.122.0 netmask 255.255.255.0 {
    option routers 192.168.122.1;
    option subnet-mask 255.255.255.0;
    range dynamic-bootp 192.168.122.50 192.168.122.200;
    default-lease-time 21600;
    max-lease-time 43200;
    next-server 192.168.122.1;
}


'''
    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
    mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
    if os.path.exists(dhcpd):
        os.remove(dhcpd)
    with os.fdopen(os.open(dhcpd, flags, mode), "w") as f:
        f.write(file)

    routers = dic["routers"]
    netmask = dic["netmask"]
    dynamic_bootp = dic["dynamicBootp"]
    subnet = dic["subnet"]
    ip_map_list = dic["ipMap"]

    ip_map = parser_ip_map(ip_map_list)
    # Add fixed IP
    for k, v in ip_map.items():
        file = f"host {v}aa {'{'}\nhardware ethernet {k};\nfixed-address {v};\n" + \
            "if option architecture-type = 00:07 or option architecture-type = 00:09 " + \
            "{\nfilename \"x86_uefi/grubx64.efi\";\n} else if option architecture-type = 00:0b " + \
            "{\nfilename \"arm_uefi/grubaa64.efi\";  } else{\nfilename \"pxelinux/pxelinux.0\";\n}}\n\n"
        flags = os.O_WRONLY | os.O_EXCL
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        with os.fdopen(os.open(dhcpd, flags, mode), "a") as f:
            f.write(file)
            f.close()
    alter(dhcpd, "192.168.122.0", subnet)
    alter(dhcpd, "255.255.255.0", netmask)
    line_break = ";\n"
    alter_after(dhcpd, "option routers ", routers + line_break)
    alter_after(dhcpd, "option subnet-mask ", netmask + line_break)
    alter_after(dhcpd, "range dynamic-bootp ", dynamic_bootp + line_break)
    alter_after(dhcpd, "next-server ", ip + line_break)
    command_syn("systemctl start dhcpd")
    command_syn("systemctl enable dhcpd")


def get_file_by_path(path):
    output = subs.check_output(['/usr/bin/ls', '%s' % path])
    output = output.decode()[:-1]
    files = output.split("\n")
    return files


def chmod_by_file(path):
    s = 'chmod -R 755 %s' % path
    command_syn(s)


def command_syn(command):
    li = command.split(' ')
    subs.run(li)


def command_link(command1, command2):
    li_1 = command1.split(' ')
    li_2 = command2.split(' ')
    p1 = subs.Popen(li_1, stdout=subs.PIPE)
    p2 = subs.Popen(li_2, stdin=p1.stdout, stdout=subs.PIPE)
    output = p2.communicate(timeout=10)[0]
    retcode = p2.returncode
    return retcode, output
